'''GFF format utilities'''
import itertools
import warnings
import os
import shlex
import sys

import HTSeq
from HTSeq._HTSeq import *
from HTSeq.utils import FileOrSequence


# GFF regular expressions for cache
_re_attr_main = re.compile("\s*([^\s\=]+)[\s=]+(.*)")
_re_attr_empty = re.compile("^\s*$")
_re_gff_meta_comment = re.compile("##\s*(\S+)\s+(\S*)")



class GenomicFeature(object):
    """A genomic feature, i.e., an interval on a genome with metadata.

    At minimum, the following information should be provided by slots:

      name: a string identifying the feature (e.g., a gene symbol)
      type: a string giving the feature type (e.g., "gene", "exon")
      iv: a GenomicInterval object specifying the feature locus
    """

    def __init__(self, name, type_, interval):
        self.name = name
        self.type = sys.intern(type_)
        self.iv = interval

    def __repr__(self):
        return "<%s: %s '%s' at %s: %d -> %d (strand '%s')>" % \
            (self.__class__.__name__, self.type, self.name,
             self.iv.chrom, self.iv.start_d, self.iv.end_d, self.iv.strand)

    def __eq__(self, other):
        if not isinstance(other, GenomicFeature):
            return False
        return self.name == other.name and self.type == other.type and \
            self.iv == other.iv

    def __neq__(self, other):
        if not isinstance(other, GenomicFeature):
            return True
        return not self.__eq__(other)

    def __hash__(self):
        return (self.name, self.type, self.iv).__hash__()

    def get_gff_line(self, with_equal_sign=False):
        try:
            source = self.source
        except AttributeError:
            source = "."
        try:
            score = self.score
        except AttributeError:
            score = "."
        try:
            frame = self.frame
        except AttributeError:
            frame = "."
        try:
            attr = self.attr
        except AttributeError:
            attr = {'ID': self.name}
        if with_equal_sign:
            sep = "="
        else:
            sep = " "
        attr_str = '; '.join(
            ['%s%s\"%s\"' % (ak, sep, attr[ak]) for ak in attr])
        return "\t".join(str(a) for a in (self.iv.chrom, source,
                         self.type, self.iv.start + 1, self.iv.end, score,
                         self.iv.strand, frame, attr_str)) + "\n"







class GFF_Reader(FileOrSequence):
    """Parse a GFF file

    Pass the constructor either a file name or an iterator of lines of a
    GFF files. If a file name is specified, it may refer to a gzip compressed
    file.

    Iterating over the object then yields GenomicFeature objects.

    Args:
        filename_or_sequence: input file or iterator of lines
        end_included: whether the end coordinate of intervals is included in
          the interval itself. This is common in GTF but not the Python
          standard, hence this argument.
        gff_version: Which version of the GFF format to use (2 or 3). The None
          default has the following meaning. If the input is a filename, use
          version 2 if it ends with gtf or gtf.gz (case insensitive), else use
          version 3. If an iterator, use version 2 by default. Notice that GFF3
          does not use quotes in gene names et similia, while GTF does.
    """

    def __init__(self, filename_or_sequence, end_included=True, gff_version=None):
        super().__init__(filename_or_sequence)
        self.end_included = end_included
        self.metadata = {}
        if gff_version is None:
            self._guess_gff_version()

    def _guess_gff_version(self):
        if not self.fos_is_path:
            gff_version = 2
        else:
            fos = os.fspath(self.fos)
            if fos.lower().endswith((".gtf.gz", ".gtf.gzip", ".gtf")):
                gff_version = 2
            else:
                gff_version = 3
        self.gff_version = gff_version

    def __iter__(self):
        for line in super().__iter__():
            if isinstance(line, bytes):
                line = line.decode()
            if line == "\n":
                continue
            if line.startswith('#'):
                if line.startswith("##"):
                    mo = _re_gff_meta_comment.match(line)
                    if mo:
                        self.metadata[mo.group(1)] = mo.group(2)
                continue
            (seqname, source, feature, start, end, score,
             strand, frame, attributeStr) = line.split("\t", 8)
            (attr, name) = self.parse_GFF_attribute_string(
                    attributeStr,
                    True,
                    self.gff_version,
                    )
            iv = GenomicInterval(
                    seqname,
                    int(start) - 1, int(end) - 1 + int(self.end_included),
                    strand)
            f = GenomicFeature(name, feature, iv)
            if score != ".":
                score = float(score)
            if frame != ".":
                frame = int(frame)
            f.source = source
            f.score = score
            f.frame = frame
            f.attr = attr
            yield f

    @staticmethod
    def parse_GFF_attribute_string(
            attrStr,
            extra_return_first_value=False,
            gff_version=2,
        ):
        """Parses a GFF attribute string and returns it as a dictionary.
    
        If 'extra_return_first_value' is set, a pair is returned: the dictionary
        and the value of the first attribute. This might be useful if this is the
        ID.

        Args:
            attrStr: the GFF attribute string to parse
            extra_return_first_value: whether to return the pair explained above
            gff_version: which GFF format rules to use (2 or 3)
        """
        if attrStr.endswith("\n"):
            attrStr = attrStr[:-1]
        d = {}
        first_val = "_unnamed_"

        if gff_version == 2:
            iterator = quotesafe_split(attrStr.encode())
        else:
            # GFF3 does not care about quotes
            iterator = attrStr.encode().split(b';')

        for (i, attr) in enumerate(iterator):
            attr = attr.decode()
            if _re_attr_empty.match(attr):
                continue

            if (gff_version == 2) and (attr.count('"') not in (0, 2)):
                raise ValueError(
                    "The attribute string seems to contain mismatched quotes.")
            mo = _re_attr_main.match(attr)
            if not mo:
                raise ValueError("Failure parsing GFF attribute line")
            val = mo.group(2)

            # GFF3 does not split quotes
            if (gff_version == 2) and val.startswith('"') and val.endswith('"'):
                val = val[1:-1]
            d[sys.intern(mo.group(1))] = sys.intern(val)
            if extra_return_first_value and i == 0:
                first_val = val
        if extra_return_first_value:
            return (d, first_val)
        else:
            return d


def _parse_feature_query(feature_query):
    if '"' not in feature_query:
        raise ValueError('Invalid feature query')
    if '==' not in feature_query:
        raise ValueError('Invalid feature query')

    idx_quote1 = feature_query.find('"')
    idx_quote2 = feature_query.rfind('"')
    attr_name = feature_query[idx_quote1+1: idx_quote2]

    idx_equal = feature_query[:idx_quote1].find('==')
    attr_cat = feature_query[:idx_equal].strip()

    return {
        'attr_cat': attr_cat,
        'attr_name': attr_name,
        }


def make_feature_dict(
        feature_sequence,
        feature_type=None,
        feature_query=None,
        ):
    """Organize a sequence of Feature objects into a nested dictionary.

    Args:
        feature_sequence (iterable of Feature): A sequence of features, e.g. as
            obtained from GFF_reader('myfile.gtf')
        feature_type (string, sequence of strings, or None): If None, collect
            all features. If a string, restrict to only one type of features,
            e.g. 'exon' (this is the most common situation). If a sequence of
            strings, restrict to the types found in the sequence, e.g.
            ['gene', 'pseudogene']. Using a feature of strings is an uncommon
            need and can lead to a higher number of ambiguous alignments: only
            use if you know what you are doing. Even then, beware that this
            option is designed to work for feature types that are "peers" and
            not obviously overlapping, such as genes and pseudogenes. If you
            select nested features types (e.g. "gene" and "exon"), you are
            likely to end up with meaningless numbers.
        feature_query (string or None): If None, all features of the selected
            types will be collected. If a string, it has to be in the format:

        <feature_attribute> == <attr_value>

        e.g.

        'gene_id == "Fn1"'

        (note the double quotes inside).

        Then only that feature will be collected. Using this argument is more
        efficient than collecting all features and then pruning it down to a
        single one.

    Returns:
        dict with all the feature types as keys. Each value is again a dict,
        now of feature names. The values of this dict is a list of features.

    Example: Let's say you load the C. elegans GTF file from Ensembl and make a
    feature dict:

    >>> gff = HTSeq.GFF_Reader("Caenorhabditis_elegans.WS200.55.gtf.gz")
    >>> worm_features_dict = HTSeq.make_feature_dict(gff)

    (This command may take a few minutes to deal with the 430,000 features
    in the GTF file. Note that you may need a lot of RAM if you have millions
    of features.)

    Then, you can simply access, say, exon 0 of gene "F08E10.4" as follows:
    >>> worm_features_dict['exon']['F08E10.4'][0]
    <GenomicFeature: exon 'F08E10.4' at V: 17479353 -> 17479001 (strand '-')>
    """

    if feature_query is not None:
        feature_qdic = _parse_feature_query(feature_query)

    features = {}
    for f in feature_sequence:
        if any(ft in (None, f.type) for ft in feature_type):
            if f.type not in features:
                features[f.type] = {}
            res_ftype = features[f.type]

            if feature_query is not None:
                # Skip the features that don't even have the right attr
                if feature_qdic['attr_cat'] not in f.attr:
                    continue
                # Skip the ones with an attribute with a different name
                # from the query (e.g. other genes)
                if f.attr[feature_qdic['attr_cat']] != feature_qdic['attr_name']:
                    continue

            if f.name not in res_ftype:
                res_ftype[f.name] = [f]
            else:
                res_ftype[f.name].append(f)
    return features


def make_feature_genomicarrayofsets(
        feature_sequence,
        id_attribute,
        feature_type=None,
        feature_query=None,
        additional_attributes=None,
        stranded=False,
        verbose=False,
        add_chromosome_info=False,
        ):
    """Organize a sequence of Feature objects into a GenomicArrayOfSets.

    Args:
        feature_sequence (iterable of Feature): A sequence of features, e.g. as
            obtained from GFF_reader('myfile.gtf')
        id_attribute (string or sequence of strings): An attribute to use to
            identify the feature in the output data structures (e.g.
            'gene_id'). If this is a list, the combination of all those
            attributes, separated by colons (:), will be used as an identifier.
            For instance, ['gene_id', 'exon_number'] uniquely identifies
            specific exons.
        feature_type (string, sequence of strings, or None): If None, collect
            all features. If a string, restrict to only one type of features,
            e.g. 'exon'. If a sequence of strings, restrict to the types found
            in the sequence, e.g. 'gene' and 'pseudogene'
        feature_query (string or None): If None, all features of the selected
            types will be collected. If a string, it has to be in the format:

        <feature_attribute> == <attr_value>

        e.g.

        'gene_id == "Fn1"'

        (note the double quotes inside).

        Then only that feature will be collected. Using this argument is more
        efficient than collecting all features and then pruning it down to a
        single one.

        additional_attributes (list or None): A list of additional attributes
            to be collected into a separate dict for the same features, for
            instance ['gene_name']
        stranded (bool): Whether to keep strandedness information
        verbose (bool): Whether to output progress and error messages
        add_chromosome_info (bool): Whether to add chromosome information for
            each feature. If this option is True, the fuction appends at the
            end of the "additional_attributes" list a "Chromosome" attribute.

    Returns:
        dict with two keys, 'features' with the GenomicArrayOfSets populated
        with the features, and 'attributes' which is itself a dict with
        the id_attribute as keys and the additional attributes as values.

    Example: Let's say you load the C. elegans GTF file from Ensembl and make a
    feature dict:

    >>> gff = HTSeq.GFF_Reader("Caenorhabditis_elegans.WS200.55.gtf.gz")
    >>> worm_features = HTSeq.make_feature_genomicarrayofsets(gff)

    (This command may take a few minutes to deal with the 430,000 features
    in the GTF file. Note that you may need a lot of RAM if you have millions
    of features.)

    This function is related but distinct from HTSeq.make_feature_dict. This
    function is used in htseq-count and its barcoded twin to count gene
    expression because the output GenomicArrayofSets is very efficient. You
    can use it in performance-critical scans of GFF files.
    """

    def get_id_attr(f, id_attribute):
        '''Get feature id with a single or multiple attributes'''
        if isinstance(id_attribute, str):
            try:
                feature_id = f.attr[id_attribute]
            except KeyError:
                raise ValueError(
                        "Feature %s does not contain a '%s' attribute" %
                        (f.name, id_attribute))
        else:
            feature_id = []
            for id_attr in id_attribute:
                try:
                    feature_id.append(f.attr[id_attr])
                except KeyError:
                    raise ValueError(
                            "Feature %s does not contain a '%s' attribute" %
                            (f.name, id_attr))
            feature_id = ':'.join(feature_id)
        return feature_id

    if additional_attributes is None:
        additional_attributes = []

    if feature_query is not None:
        feature_qdic = _parse_feature_query(feature_query)

    features = HTSeq.GenomicArrayOfSets("auto", stranded)
    attributes = {}
    i = 0
    try:
        for f in feature_sequence:
            if any(ft in (None, f.type) for ft in feature_type):
                feature_id = get_id_attr(f, id_attribute)

                if stranded and f.iv.strand == ".":
                    raise ValueError(
                            "Feature %s at %s does not have strand information but you are "
                            "using stranded mode. Try with unstrnded mode." %
                            (f.name, f.iv))

                if feature_query is not None:
                    # Skip the features that don't even have the right attr
                    if feature_qdic['attr_cat'] not in f.attr:
                        continue
                    # Skip the ones with an attribute with a different name
                    # from the query (e.g. other genes)
                    if f.attr[feature_qdic['attr_cat']] != feature_qdic['attr_name']:
                        continue

                features[f.iv] += feature_id
                attributes[feature_id] = [
                        f.attr[attr] if attr in f.attr else ''
                        for attr in additional_attributes]
                if add_chromosome_info:
                    attributes[feature_id] += [f.iv.chrom]

            i += 1
            if i % 100000 == 0 and verbose:
                if hasattr(feature_sequence, 'get_line_number_string'):
                    msg = "{:d} GFF lines processed.".format(i)
                else:
                    msg = "{:d} features processed.".format(i)
                sys.stderr.write(msg+'\n')
                sys.stderr.flush()
    except(KeyError, ValueError):
        if verbose:
            if hasattr(feature_sequence, 'get_line_number_string'):
                msg = "Error processing GFF file ({:}):".format(
                    feature_sequence.get_line_number_string())
            else:
                msg = "Error processing feature sequence ({:}):".format(
                    str(i+1))
            sys.stderr.write(msg+'\n')
        raise

    if verbose:
        if hasattr(feature_sequence, 'get_line_number_string'):
            msg = "{:d} GFF lines processed.".format(i)
        else:
            msg = "{:d} features processed.".format(i)
        sys.stderr.write(msg+"\n")
        sys.stderr.flush()

    if add_chromosome_info:
        additional_attributes.append('Chromosome')

    return {
        'features': features,
        'attributes': attributes,
        }



import pytest
import numpy as np
import sys
import os
import glob
import sysconfig
from pathlib import Path
import unittest
import pytest
import conftest

build_dir = "build/lib.%s-%s" % (sysconfig.get_platform(), sys.version[0:3])

sys.path.insert(0, os.path.join(os.getcwd(), build_dir))
import HTSeq


data_folder = conftest.get_data_folder()


class TestStretchVector(unittest.TestCase):
    def test_init(self):
        sv = HTSeq.StretchVector(typecode='d')
        self.assertEqual(sv.ivs, [])
        self.assertEqual(sv.stretches, [])

    def test_setitem_number(self):
        sv = HTSeq.StretchVector(typecode='d')

        # Set initial stretch
        sv[560] = 4.5
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            np.ones(1, np.float32) * 4.5,
        )

        # Overwrite
        sv[560] = 4
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            np.ones(1, np.float32) * 4,
        )

    def test_setitem_slice(self):
        sv = HTSeq.StretchVector(typecode='d')

        # Set initial stretch
        sv[100: 300] = 4.5
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            np.ones(200, np.float32) * 4.5,
        )

        # Set overlapping stretch
        sv[50:250] = 3
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        np.testing.assert_almost_equal(
            sv.stretches[0][:200],
            np.ones(200, np.float32) * 3,
        )
        np.testing.assert_almost_equal(
            sv.stretches[0][200:],
            np.ones(50, np.float32) * 4.5,
        )

        # Set new stretch
        sv[400: 450] = np.arange(50)
        self.assertEqual(len(sv.ivs), 2)
        self.assertEqual(len(sv.stretches), 2)
        np.testing.assert_almost_equal(
            sv.stretches[1],
            np.arange(50).astype(np.float32),
        )

        # Set overlapping stretch
        sv[430: 450] = np.arange(20)
        self.assertEqual(len(sv.ivs), 2)
        self.assertEqual(len(sv.stretches), 2)
        np.testing.assert_almost_equal(
            sv.stretches[1][-20:],
            np.arange(20).astype(np.float32),
        )

    def test_getitem_number(self):
        sv = HTSeq.StretchVector(typecode='d')

        # Set initial stretch
        sv[560] = 4.5
        self.assertEqual(sv[560], 4.5)
        self.assertEqual(sv[580], None)

        sv[400: 450] = np.arange(50)
        res = sv[350: 430]
        self.assertEqual(len(res.ivs), 1)
        self.assertEqual(len(res.stretches), 1)
        np.testing.assert_almost_equal(
            res.stretches[0],
            np.arange(30).astype(np.float32),
        )

    def test_todense(self):
        sv = HTSeq.StretchVector(typecode='d')
        sv[450: 455] = 6.7
        sv[460: 465] = 1.7
        res = sv.todense()
        np.testing.assert_almost_equal(
            res,
            np.array([6.7] * 5 + [np.nan] * 5 + [1.7] * 5).astype(np.float32),
        )

    def test_from_dense(self):
        array = np.empty(20, np.float32)
        array[:] = np.nan

        # All nans
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 0)
        self.assertEqual(len(sv.stretches), 0)

        # All good
        array[:] = 78
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)

        # Single flips
        array[:5] = np.nan
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        self.assertEqual(sv.ivs[0].start, 305)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            78 * np.ones(15).astype(np.float32),
        )

        array[:] = 78
        array[15:] = np.nan
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        self.assertEqual(sv.ivs[0].start, 300)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            78 * np.ones(15).astype(np.float32),
        )

        # Double flip at edges
        array[:] = 78
        array[:4] = np.nan
        array[15:] = np.nan
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 1)
        self.assertEqual(len(sv.stretches), 1)
        self.assertEqual(sv.ivs[0].start, 304)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            78 * np.ones(11).astype(np.float32),
        )

        # Double flip in the middle
        array[:] = 78
        array[2:4] = np.nan
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 2)
        self.assertEqual(len(sv.stretches), 2)
        self.assertEqual(sv.ivs[0].start, 300)
        self.assertEqual(sv.ivs[0].end, 302)
        self.assertEqual(sv.ivs[1].start, 304)
        self.assertEqual(sv.ivs[1].end, 320)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            78 * np.ones(2).astype(np.float32),
        )
        np.testing.assert_almost_equal(
            sv.stretches[1],
            78 * np.ones(16).astype(np.float32),
        )

        # Multiple flips
        array[:] = 78
        array[2:4] = np.nan
        array[8:11] = np.nan
        array[17:19] = np.nan
        sv = HTSeq.StretchVector.from_dense(array, offset=300)
        self.assertEqual(len(sv.ivs), 4)
        self.assertEqual(len(sv.stretches), 4)
        self.assertEqual(sv.ivs[0].start, 300)
        self.assertEqual(sv.ivs[0].end, 302)
        self.assertEqual(sv.ivs[1].start, 304)
        self.assertEqual(sv.ivs[1].end, 308)
        self.assertEqual(sv.ivs[2].start, 311)
        self.assertEqual(sv.ivs[2].end, 317)
        self.assertEqual(sv.ivs[3].start, 319)
        self.assertEqual(sv.ivs[3].end, 320)
        np.testing.assert_almost_equal(
            sv.stretches[0],
            78 * np.ones(2).astype(np.float32),
        )
        np.testing.assert_almost_equal(
            sv.stretches[1],
            78 * np.ones(4).astype(np.float32),
        )
        np.testing.assert_almost_equal(
            sv.stretches[2],
            78 * np.ones(6).astype(np.float32),
        )
        np.testing.assert_almost_equal(
            sv.stretches[3],
            78 * np.ones(1).astype(np.float32),
        )


if __name__ == '__main__':

    suite = TestStretchVector()
    suite.test_init()



import HTSeq
import numpy
from matplotlib import pyplot as plt


sortedbamfile = HTSeq.BAM_Reader("SRR001432_head_sorted.bam")
gtffile = HTSeq.GFF_Reader("Homo_sapiens.GRCh37.56_chrom1.gtf")
halfwinwidth = 3000
fragmentsize = 200

tsspos = set()
for feature in gtffile:
    if feature.type == "exon" and feature.attr["exon_number"] == "1":
        tsspos.add(feature.iv.start_d_as_pos)

profile = numpy.zeros(2 * halfwinwidth, dtype="i")
for p in tsspos:
    window = HTSeq.GenomicInterval(
        p.chrom,
        p.pos - halfwinwidth - fragmentsize,
        p.pos + halfwinwidth + fragmentsize,
        ".",
    )
    for almnt in sortedbamfile[window]:
        almnt.iv.length = fragmentsize
        if p.strand == "+":
            start_in_window = almnt.iv.start - p.pos + halfwinwidth
            end_in_window = almnt.iv.end - p.pos + halfwinwidth
        else:
            start_in_window = p.pos + halfwinwidth - almnt.iv.end
            end_in_window = p.pos + halfwinwidth - almnt.iv.start
        start_in_window = max(start_in_window, 0)
        end_in_window = min(end_in_window, 2 * halfwinwidth)
        if start_in_window >= 2 * halfwinwidth or end_in_window < 0:
            continue
        profile[start_in_window:end_in_window] += 1



import os
import sys
import sysconfig
import numpy

build_dir = "build/lib.%s-%s" % (sysconfig.get_platform(), sys.version[0:3])

sys.path.insert(0, os.path.join(os.getcwd(), build_dir))
import HTSeq
os.chdir("example_data")

print("TSS test, scheme 1")
exec(compile(open(os.path.join("..", "doc", "code_examples", "tss1.py")).read(), os.path.join("..", "doc", "tss1.py"), 'exec'))
profile1 = profile.copy()
print("finished")

print("TSS test, scheme 2")
exec(compile(open(os.path.join("..", "doc", "code_examples", "tss2.py")).read(), os.path.join("..", "doc", "tss2.py"), 'exec'))
profile2 = profile.copy()
print("finished", end=' ')

if (profile2 == profile1).all():
    print("and matches result from scheme 1.")
else:
    print("and differs to result from scheme 1!  <<<!!!>>>")

print("TSS test, scheme 3")
exec(compile(open(os.path.join("..", "doc", "code_examples", "tss3.py")).read(), os.path.join("..", "doc", "tss3.py"), 'exec'))
profile3 = profile.copy()
print("finished", end=' ')

if (profile3 == profile1).all():
    print("and matches result from scheme 1.")
else:
    print("and differs to result from scheme 1!  <<<!!!>>>")



import HTSeq
import sys


class ReadsStatistics(object):
    """
    For storing the read counts
    """

    def __init__(self, feature_attr, read_io_object):
        '''Initialize storage for the read counts

        Args:
            feature_attr (list of str): Feature names
            read_io_object (ReadsIO): Read processor
        '''
        self.read_io_obj = read_io_object

        self.empty = 0
        self.ambiguous = 0
        self.notaligned = 0
        self.lowqual = 0
        self.nonunique = 0
        self.num_reads_processed = 0
        self.counts = {key: 0 for key in feature_attr}

    def add_num_reads_processed(self):
        self.num_reads_processed += 1

    def add_empty_read(self, read_sequence):
        self.empty += 1
        self.read_io_obj.write_to_samout(read_sequence, "__no_feature")

    def add_ambiguous_read(self, read_sequence, assignment):
        self.ambiguous += 1
        self.read_io_obj.write_to_samout(read_sequence, assignment)

    def add_low_quality_read(self, read_sequence):
        self.lowqual += 1
        self.read_io_obj.write_to_samout(read_sequence, "__too_low_aQual")

    def add_not_unique_read(self, read_sequence):
        self.nonunique += 1
        self.read_io_obj.write_to_samout(read_sequence, "__alignment_not_unique")

    def add_not_aligned_read(self, read_sequence):
        self.notaligned += 1
        self.read_io_obj.write_to_samout(read_sequence, "__not_aligned")

    def add_to_count(self, feature, value=1):
        self.counts[feature] += value

    def add_good_read_assignment(self, read_sequence, assignment):
        self.read_io_obj.write_to_samout(read_sequence, assignment)

    def print_progress(self, force_print=False):
        """
        Simple function to update the progress of reads processing.
        Parameters
        ----------
        force_print : boolean
            Whether to print progress regardless of number of reads processed
            or not.
        """

        if force_print:
            do_print = True
        else:
            do_print = (
                self.num_reads_processed > 0 and self.num_reads_processed % 100000 == 0
            )

        if do_print:
            sys.stderr.write(
                "%d alignment record%s processed.\n"
                % (
                    self.num_reads_processed,
                    "s" if not self.read_io_obj.pe_mode else " pairs",
                )
            )
            sys.stderr.flush()

    def get_output(self, isam):
        return {
            "isam": isam,
            "counts": self.counts,
            "empty": self.empty,
            "ambiguous": self.ambiguous,
            "lowqual": self.lowqual,
            "notaligned": self.notaligned,
            "nonunique": self.nonunique,
        }



import itertools
import pysam
import HTSeq
import sys


class ReadsIO(object):
    """docstring for ReadsIO."""

    def __init__(
        self,
        sam_filename,
        samout_filename,
        samout_format,
        supplementary_alignment_mode,
        secondary_alignment_mode,
        order,
        max_buffer_size,
    ):

        # Set by _prepare_bam_sam_file_parser function below.
        self.pe_mode = None
        self.read_seq = None
        self.read_seq_file = None
        self.template = None
        self.samoutfile = None
        self.samout_format = samout_format

        self._set_BAM_reader(sam_filename)
        self._set_output_template(samout_filename, samout_format)
        self._set_read_seq(
            supplementary_alignment_mode,
            secondary_alignment_mode,
            order,
            max_buffer_size,
        )

    def write_to_samout(self, read_sequence, assignment):
        if self.samoutfile is None:
            return
        if not self.pe_mode:
            # TODO not sure if this is good in all honesty..
            read_sequence = (read_sequence,)
        for read in read_sequence:
            if read is not None:
                read.optional_fields.append(("XF", assignment))
                if self.template is not None:
                    self.samoutfile.write(read.to_pysam_AlignedSegment(self.template))
                elif self.samout_format in ("SAM", "sam"):
                    self.samoutfile.write(read.get_sam_line() + "\n")
                else:
                    raise ValueError(
                        "BAM/SAM output: no template and not a test SAM file",
                    )

    def close_samoutfile(self):
        if self.samoutfile is not None:
            self.samoutfile.close()

    def _set_BAM_reader(self, sam_filename):
        """
        Convert the input SAM/BAM files into a parser.

        Parameters
        ----------
        sam_filename : str
            The name of SAM/BAM file to write out all SAM alignment records into.

        """
        if sam_filename == "-":
            self.read_seq_file = HTSeq.BAM_Reader(sys.stdin)
        else:
            self.read_seq_file = HTSeq.BAM_Reader(sam_filename)

    def get_chromosome_names_header(self):
        """ Reads BAM header and returns a list of contigs, or None if no SQ in header """
        contigs = None
        sq = self.read_seq_file.get_header_dict().get("SQ")
        if sq is not None:
            contigs = []
            for sq_record in sq:
                sn = sq_record.get("SN")
                if sn:
                    contigs.append(sn)
        return contigs

    def _set_read_seq(
        self,
        supplementary_alignment_mode,
        secondary_alignment_mode,
        order,
        max_buffer_size,
    ):

        """
        Prepare the BAM/SAM file iterator.
        Note, only run this after _set_BAM_reader as you need self.read_seq_file to be set.
        This will create a parser and prepare an iterator for it.
        Depending on whether we have paired-end reads or not, different iterator
        will be returned.

        Parameters
        ----------
        supplementary_alignment_mode : str
            Whether to score supplementary alignments (0x800 flag).
            Choices: score or ignore.
        secondary_alignment_mode : str
            Whether to score secondary alignments (0x100 flag).
            Choices: score or ignore.
        order : str
            Can only be either 'pos' or 'name'. Sorting order of <alignment_file>.
        max_buffer_size : int
            When <alignment_file> is paired end sorted by position, allow only so many reads to stay in memory
            until the mates are found (raising this number will use more memory).
            Has no effect for single end or paired end sorted by name.

        """

        read_seq_iter = iter(self.read_seq_file)
        # Catch empty BAM files
        try:
            first_read = next(read_seq_iter)
            self.pe_mode = first_read.paired_end
        # FIXME: catchall can hide subtle bugs
        except:
            first_read = None
            self.pe_mode = False
        if first_read is not None:
            self.read_seq = itertools.chain([first_read], read_seq_iter)
        else:
            self.read_seq = []

        if self.pe_mode:
            if (supplementary_alignment_mode == "ignore") and (
                secondary_alignment_mode == "ignore"
            ):
                primary_only = True
            else:
                primary_only = False
            if order == "name":
                self.read_seq = HTSeq.pair_SAM_alignments(
                    self.read_seq, primary_only=primary_only
                )
            elif order == "pos":
                self.read_seq = HTSeq.pair_SAM_alignments_with_buffer(
                    self.read_seq,
                    max_buffer_size=max_buffer_size,
                    primary_only=primary_only,
                )
            else:
                raise ValueError("Illegal order specified.")

    def _set_output_template(self, samout_filename, samout_format):
        """
        Set up the SAM/BAM output files (and corresponding template) if possible.

        Parameters
        ----------
        samout_filename : str
            The name of SAM/BAM file to write out all SAM alignment records into.
        samout_format : str
            Format of the output files denoted by samouts.
            Choices: SAM, BAM, sam, bam.

        """
        if samout_filename is None:
            self.template = None
            self.samoutfile = None
        elif samout_format in ("bam", "BAM"):
            self.template = self.read_seq_file.get_template()
            self.samoutfile = pysam.AlignmentFile(
                samout_filename,
                "wb",
                template=self.template,
            )
        elif (samout_format in ("sam", "SAM")) and hasattr(
            self.read_seq_file, "get_template"
        ):
            self.template = self.read_seq_file.get_template()
            self.samoutfile = pysam.AlignmentFile(
                samout_filename,
                "w",
                template=self.template,
            )
        else:
            self.template = None
            self.samoutfile = open(samout_filename, "w")



import pytest
import sys
import os
import glob
import sysconfig
from pathlib import Path

build_dir = "build/lib.%s-%s" % (sysconfig.get_platform(), sys.version[0:3])

sys.path.insert(0, os.path.join(os.getcwd(), build_dir))
import HTSeq


def test_fasta_parser(data_folder):
    print("Test Fasta parser")
    for seq in HTSeq.FastaReader(data_folder+'fastaExLong.fa'):
        pass
    print("Test passed")
    print("Test Fasta parser (raw iterator)")
    for seq in HTSeq.FastaReader(data_folder+'fastaExLong.fa',
                                 raw_iterator=True):
        pass
    print("Test passed")

    print('Test Fasta parser (with statement)')
    with HTSeq.FastaReader(data_folder+'fastaExLong.fa') as f:
        for seq in f:
            pass
    print('Test passed')

    print('Test Fasta parser (with statement and file handle)')
    with open(data_folder+'fastaExLong.fa') as fraw:
        f = HTSeq.FastaReader(fraw)
        for seq in f:
            pass
    print('Test passed')


def test_fastq_parser(data_folder):
    print("Test Fastq parser")
    for seq in HTSeq.FastqReader(data_folder+'fastqEx.fastq'):
        pass
    print("Test passed")
    print("Test Fastq parser on gzip input")
    for seq in HTSeq.FastqReader(data_folder+'fastqExgzip.fastq.gz'):
        pass
    print("Test passed")
    print("Test Fastq parser on gzip input (raw iterator)")
    for seq in HTSeq.FastqReader(data_folder+'fastqExgzip.fastq.gz',
                                 raw_iterator=True):
        pass
    print("Test passed")

    print('Test Fastq parser (with statement)')
    with HTSeq.FastqReader(data_folder+'fastqExgzip.fastq.gz') as f:
        for seq in f:
            pass
    print('Test passed')

    print('Test Fastq parser (with statement and file handle)')
    import gzip
    with gzip.open(data_folder+'fastqExgzip.fastq.gz', 'rt') as fraw:
        f = HTSeq.FastqReader(fraw)
        for seq in f:
            pass
    print('Test passed')


def test_path_like_reading(data_folder):
    data_file = Path(data_folder) / 'yeast_RNASeq_excerpt.sam'
    print('Test BAM reader with pathlib.Path')
    bamfile = HTSeq.BAM_Reader(data_file)
    for read in bamfile:
        pass
    print('Test passed')

    print('Test BAM reader (with statement) with pathlib.Path')
    with HTSeq.BAM_Reader(data_file) as f:
        for read in f:
            pass
    print('Test passed')


def test_open_handle_reading(data_folder):
    data_file = Path(data_folder) / 'yeast_RNASeq_excerpt.sam'
    print('Test BAM reader with open handle')
    with open(data_file) as f:
        reader = HTSeq.BAM_Reader(f)
        for read in reader:
            pass
    print('Test passed')


def test_unreadable_object_reading():
    unreadable_object = object()

    print('Test BAM reader with unreadable object')
    with pytest.raises(TypeError):
        reader = HTSeq.BAM_Reader(unreadable_object)
        for read in reader:
            pass
    print('Test passed')

    print('Test BAM reader with unreadable object, context manager')
    with pytest.raises(TypeError):
        with HTSeq.BAM_Reader(unreadable_object) as reader:
            for read in reader:
                pass
    print('Test passed')


def test_bam_reader(data_folder):
    print('Test BAM reader')
    bamfile = HTSeq.BAM_Reader(data_folder+"yeast_RNASeq_excerpt.sam")
    for read in bamfile:
        pass
    print('Test passed')

    print('Test BAM reader (with statement)')
    with HTSeq.BAM_Reader(data_folder+"yeast_RNASeq_excerpt.sam") as f:
        for read in f:
            pass
    print('Test passed')


def test_bam_inconsistent_mate(data_folder):
    print('Test inconsistent BAM file')
    bamfile = HTSeq.BAM_Reader(data_folder+"inconsistent_mate.bam")
    for read in bamfile:
        pass
    print("Test passed")


def test_bam_optional_field(data_folder):
    print('Test optional fields in BAM alignment')
    bamfile = HTSeq.BAM_Reader(data_folder+"inconsistent_mate.bam")
    for read in bamfile:
        read.has_optional_field('NO')
        break
    print("Test passed")


def test_GFF3(data_folder):
    print('Test GFF3 reader')
    with HTSeq.GFF_Reader(data_folder+'GCF_000001405.39_GRCh38.p13_genomic_subsample.gff.gz') as reader:
        for line in reader:
            pass
    print("Test passed")


def test_pickle():
    import pickle

    print('Test pickling and inpickling')
    pickles = [
            {'name': 'HTSeq.Sequence',
             'object': HTSeq.Sequence(b'ACTG', 'sequence'),
             'assert_properties': ('seq', 'name', 'descr')},
            ]

    for pic in pickles:
        print('Pickling '+pic['name'])
        pickled = pickle.dumps(pic['object'])
        print('Done')

        print('Unpickling '+pic['name'])
        unpick = pickle.loads(pickled)
        print('Done')

        if 'assert_properties' in pic:
            print('Checking serialized/deserialized')
            for prop in pic['assert_properties']:
                assert getattr(pic['object'], prop) == getattr(unpick, prop)
            print('Done')
    print("Test passed")


def test_bamfile_nosq(data_folder):
    print('Test parsing BAM file with no SQ field (e.g. PacBio)')
    bamfile = HTSeq.BAM_Reader(
            data_folder+"short_test_ccs.bam",
            check_sq=False)
    for read in bamfile:
        pass
    print("Test passed")



import glob

counter_code = '''
<!-- Start of StatCounter Code -->
<script type="text/javascript">
var sc_project=5680199; 
var sc_invisible=1; 
var sc_partition=63; 
var sc_click_stat=1; 
var sc_security="6e579550"; 
</script>

<script type="text/javascript"
src="http://www.statcounter.com/counter/counter.js"></script><noscript><div
class="statcounter"><a title="counter for iweb"
href="http://www.statcounter.com/iweb/" target="_blank"><img
class="statcounter"
src="http://c.statcounter.com/5680199/0/6e579550/1/"
alt="counter for iweb" ></a></div></noscript>
<!-- End of StatCounter Code -->
'''

def process_file(fn):
   lines = open(fn).readlines()
   for i in range(len(lines)):
      if lines[i].count('statcounter') > 0:
         return
   for i in range(len(lines)):
      if lines[i].count('<\body>') > 0:
         break
   lines.insert(i, counter_code)
   f = open(fn, 'w')
   for l in lines:
      f.write(l)
   f.close()
      
for fn in glob.glob("_build/html/*html"):
   process_file(fn)      

   



import os
import pytest


def get_data_folder():
    cwd = os.getcwd().rstrip('/')
    if cwd.endswith('example_data'):
        return cwd+'/'
    else:
        return cwd+'/example_data/'


def get_docs_folder():
    cwd = os.getcwd().rstrip('/')
    if cwd.endswith('doc'):
        return cwd+'/'
    elif cwd.endswith('example_data'):
        return cwd[:-len('example_data')]+'doc/'
    else:
        return cwd+'/doc/'


# Same as fixtures
@pytest.fixture(scope="module")
def data_folder():
    return get_data_folder()


@pytest.fixture(scope="module")
def docs_folder():
    return get_docs_folder()






#!/usr/bin/env python
import os
import sys
from setuptools import setup, Command, Extension
from setuptools.command.build_py import build_py
import numpy


this_directory = os.path.abspath(os.path.dirname(__file__))


def update_version():
    import subprocess as sp
    # Try getting the latest release tag
    version = None
    try:
        output = sp.check_output(
                'git describe --tags --abbrev=0',
                shell=True,
                )
        if not isinstance(output, str):
            output = output.decode().strip('\n')
        if output.startswith('release_'):
            version = output.split('_')[1]
        print('VERSION updated: '+version)
    except:
        pass

    # Fallback is reading from VERSION file
    if version is not None:
        with open(os.path.join(this_directory, 'VERSION'), 'wt') as fversion:
            fversion.write(version+'\n')
    else:
        with open(os.path.join(this_directory, 'VERSION')) as fversion:
            version = fversion.readline().rstrip()

    # Update version from VERSION file into module
    with open(os.path.join(this_directory, 'HTSeq', '_version.py'), 'wt') as fversion:
        fversion.write('__version__ = "'+version+'"')

    return version


if ((sys.version_info[0] == 2) or
   (sys.version_info[0] == 3 and sys.version_info[1] < 7)):
    sys.stderr.write("Error in setup script for HTSeq:\n")
    sys.stderr.write("HTSeq requires Python 3.7+.")
    sys.exit(1)


# Get version
version = update_version()

# Get README file content
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

# Check OS-specific quirks
# NOTE: setuptools < 18.0 has issues with Cython as a dependency
# NOTE: old setuptools < 18.0 has issues with extras
kwargs = dict(
    setup_requires=[
          'Cython',
          'numpy',
          'pysam',
    ],
    install_requires=[
        'numpy',
        'pysam',
    ],
    extras_require={
        'htseq-qa': ['matplotlib>=1.4'],
        'test': [
            'scipy>=1.5.0',
            'pytest>=6.2.5',
            'pandas>=1.1.0',
            'matplotlib>=1.4',
        ],
    },
  )

def get_library_dirs_cpp():
    '''OSX 10.14 and later messed up C/C++ library locations'''
    if sys.platform == 'darwin':
        return ['/usr/X11R6/lib']
    else:
        return []


def get_extra_args_cpp():
    '''OSX 101.14 and later refuses to use libstdc++'''
    if sys.platform == 'darwin':
        return ['-stdlib=libc++']
    else:
        return []


class Preprocess_command(Command):
    '''Cython and SWIG preprocessing'''
    description = "preprocess Cython and SWIG files for HTSeq"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.swig_and_cython()

    def swig_and_cython(self):
        import os
        from shutil import copy
        from subprocess import check_call
        from subprocess import SubprocessError

        def c(x): return check_call(x, shell=True)
        def p(x): return self.announce(x, level=2)

        # CYTHON
        p('cythonizing')
        cython = os.getenv('CYTHON', 'cython')
        try:
            c(cython+' --version')
        except SubprocessError:
            if os.path.isfile('src/_HTSeq.c'):
                p('Cython not found, but transpiled file found')
            else:
                raise
        else:
            c(cython+' -3 src/HTSeq/_HTSeq.pyx -o src/_HTSeq.c')

        # SWIG
        p('SWIGging')
        swig = os.getenv('SWIG', 'swig')
        pyswigged = 'src/StepVector.py'
        try:
            c(swig+' -Wall -c++ -python -py3 src/StepVector.i')
            p('Files transpiled')
        except SubprocessError:
            if (os.path.isfile('src/StepVector_wrap.cxx') and
                    os.path.isfile('src/StepVector.py')):
                p('SWIG not found, but transpiled files found')
            else:
                p('swig not found and traspiled files not found.\n' +
                  'Install SWIG via your package manager (linux) or ' +
                  'via "brew install swig" (OSX - via homebrew)')
                raise
        p('moving swigged .py module')
        copy(pyswigged, 'HTSeq/StepVector.py')

        p('done')


class Build_with_preprocess(build_py):
    def run(self):
        self.run_command('preprocess')
        build_py.run(self)


def lazy_numpy_include_dir():
    """Lazily obtain NumPy include directory."""
    try:
        import numpy
        return os.path.join(os.path.dirname(numpy.__file__), 'core', 'include')
    except ImportError:
        sys.stderr.write("Failed to import 'numpy'. It is required for building HTSeq.\n")
        sys.exit(1)


setup(name='HTSeq',
      version=version,
      author='Simon Anders, Fabio Zanini',
      author_email='fabio.zanini@unsw.edu.au',
      maintainer='Fabio Zanini',
      maintainer_email='fabio.zanini@unsw.edu.au',
      url='https://github.com/htseq',
      description="A framework to process and analyze data from " +
                  "high-throughput sequencing (HTS) assays",
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='GPL3',
      classifiers=[
         'Development Status :: 5 - Production/Stable',
         'Topic :: Scientific/Engineering :: Bio-Informatics',
         'Intended Audience :: Developers',
         'Intended Audience :: Science/Research',
         'License :: OSI Approved :: GNU General Public License (GPL)',
         'Operating System :: POSIX',
         'Programming Language :: Python'
      ],
      ext_modules=[
         Extension(
             'HTSeq._HTSeq',
             ['src/_HTSeq.c'],
             #include_dirs=[lazy_numpy_include_dir()],#+get_include_dirs(),
             include_dirs=[numpy.get_include()],
             extra_compile_args=['-w']),
         Extension(
             'HTSeq._StepVector',
             ['src/StepVector_wrap.cxx'],
             #include_dirs=get_include_dirs(cpp=True),
             library_dirs=get_library_dirs_cpp(),
             extra_compile_args=['-w'] + get_extra_args_cpp(),
             extra_link_args=get_extra_args_cpp(),
             ),
      ],
      py_modules=[
         'HTSeq._HTSeq_internal',
         'HTSeq.StepVector',
         'HTSeq.StretchVector',
         'HTSeq._version',
         'HTSeq.scripts.qa',
         'HTSeq.scripts.count',
         'HTSeq.scripts.count_with_barcodes',
         'HTSeq.scripts.utils',
         'HTSeq.utils',
         'HTSeq.features',
         'HTSeq.scripts.count_features.count_features_per_file',
         'HTSeq.scripts.count_features.reads_io_processor',
         'HTSeq.scripts.count_features.reads_stats',
      ],
      scripts=[
         'scripts/htseq-qa',
         'scripts/htseq-count',
         'scripts/htseq-count-barcodes',
      ],
      cmdclass={
          'preprocess': Preprocess_command,
          'build_py': Build_with_preprocess,
          },
      **kwargs
      )



import HTSeq
import numpy
from matplotlib import pyplot as plt


bamfile = HTSeq.BAM_Reader("SRR001432_head.bam")
gtffile = HTSeq.GFF_Reader("Homo_sapiens.GRCh37.56_chrom1.gtf")
halfwinwidth = 3000
fragmentsize = 200

coverage = HTSeq.GenomicArray("auto", stranded=False, typecode="i")
for almnt in bamfile:
    if almnt.aligned:
        almnt.iv.length = fragmentsize
        coverage[almnt.iv] += 1

tsspos = set()
for feature in gtffile:
    if feature.type == "exon" and feature.attr["exon_number"] == "1":
        tsspos.add(feature.iv.start_d_as_pos)

profile = numpy.zeros(2 * halfwinwidth, dtype="i")
for p in tsspos:
    window = HTSeq.GenomicInterval(
        p.chrom, p.pos - halfwinwidth, p.pos + halfwinwidth, "."
    )
    wincvg = numpy.fromiter(coverage[window], dtype="i", count=2 * halfwinwidth)
    if p.strand == "+":
        profile += wincvg
    else:
        profile += wincvg[::-1]



# Internal HTSeq functions, not part of the API
import HTSeq
import numpy


def GenomicInterval_range(gi, step):
    for pos in range(gi.start, gi.end, step):
        yield HTSeq.GenomicPosition(gi.chrom, pos, gi.strand)


def GenomicInterval_xranged(gi, step):
    if gi.strand == "-":
        step *= -1
    for pos in range(gi.start_d, gi.end_d, step):
        yield HTSeq.GenomicPosition(gi.chrom, pos, gi.strand)


def ChromVector_steps(cv):
    '''Steps over a ChromVector


    NOTE: ChromVectors use an offset which is also iv.start compared to their
          storage objects that start at 0.
    '''
    # "Steps" of an ndarray (or memmap?)-storaged ChromVector
    if isinstance(cv.array, numpy.ndarray):
        start = cv.iv.start
        prev_val = None
        for i in range(cv.iv.start, cv.iv.end):
            val = cv.array[i - cv.offset]
            if prev_val is None or val != prev_val:
                if prev_val is not None:
                    yield (HTSeq.GenomicInterval(cv.iv.chrom, start, i, cv.iv.strand), prev_val)
                prev_val = val
                start = i
        yield (HTSeq.GenomicInterval(
            cv.iv.chrom, start, cv.iv.end, cv.iv.strand), prev_val,
            )

    # Steps of a StepVector-storaged ChromVector
    elif isinstance(cv.array, HTSeq.StepVector.StepVector):
        for start, stop, value in cv.array[
                cv.iv.start - cv.offset: cv.iv.end - cv.offset].get_steps():
            yield (HTSeq.GenomicInterval(
                cv.iv.chrom, start + cv.offset, stop + cv.offset, cv.iv.strand), value,
                )

    # Steps in a StretchVector behave similar to a full numpy array, but uses
    # np.nan for None and treats each stretch as independent, of course
    # NOTE: one could optimize this by using np.diff and flips, as we have done
    # in the StretchVector methods. For now, we leave it like this because
    # the whole point of StretchVector is to be used for stretches, not steps.
    elif isinstance(cv.array, HTSeq.StretchVector):
        for iv, stretch in cv.array:
            start = cv.offset + iv.start
            prev_val = None
            for i, val in enumerate(stretch, cv.offset + iv.start):
                # Subsequent NaNs, ignore
                if (prev_val is not None) and numpy.isnan(prev_val) and numpy.isnan(val):
                    continue
                if prev_val is None or val != prev_val:
                    # Delay yield of the first item until you meet the first
                    # unequal item, i.e. until you know the end of the step
                    if prev_val is not None:
                        yield (HTSeq.GenomicInterval(
                                    cv.iv.chrom,
                                    start,
                                    i,
                                    cv.iv.strand),
                                prev_val)
                    prev_val = val
                    start = i
            yield (HTSeq.GenomicInterval(
                cv.iv.chrom, start, cv.offset + iv.end, cv.iv.strand), prev_val,
                )
    else:
        raise SystemError("Unknown array type.")


def GenomicArray_steps(ga):
    """Steps of a GenomicArray are just the chained steps of each ChromVector"""
    for chrom, chromstrand_dict in ga.chrom_vectors.items():
        for strand, chrom_vector in chromstrand_dict.items():
            for iv, val in chrom_vector.steps():
                yield iv, val
